/*
 * uart1.c : serial1 handling routines with interrupts
 *   by uratan! 2008.9.5
 */
#include "common.h"
#include "LPC214x.h"
#include "irq.h"
#include "swi.h"

/*
 * switch uart0/uart1 here, sync with the same macro in main.c
 */
#define USE_UART1   1

#if USE_UART1
# define UxDLL U1DLL
# define UxDLM U1DLM
# define UxFCR U1FCR
# define UxIER U1IER
# define UxIIR U1IIR
# define UxLCR U1LCR
# define UxLCR U1LCR
# define UxLSR U1LSR
# define UxRBR U1RBR
# define UxTHR U1THR
# define UARTx_INT  UART1_INT
#else /* USE_UART1 */
# define UxDLL U0DLL
# define UxDLM U0DLM
# define UxFCR U0FCR
# define UxIER U0IER
# define UxIIR U0IIR
# define UxLCR U0LCR
# define UxLCR U0LCR
# define UxLSR U0LSR
# define UxRBR U0RBR
# define UxTHR U0THR
# define UARTx_INT  UART0_INT
#endif /* USE_UART1 */

/*
 * register bitmaps
 */
/* UxIER: Interrupt Enable Register */
#define IER_RBR		0x01	/* receiver buffer register(RDA) + CTI */
#define IER_THRE	0x02	/* transmit holding register */
#define IER_RLS		0x04	/* RX line status */
/* UxIIR: Interrupt Identification Register */
#define IIR_PEND	0x01
#define  IIR_RLS	(0x03<<1)	/* Receive Line Status */
#define  IIR_RDA	(0x02<<1)	/* Receive Data Available */
#define  IIR_CTI	(0x06<<1)	/* Character Time-out Indicator */
#define  IIR_THRE	(0x01<<1)	/* THRE interrupt */
/* UxLSR: Line Status Register */
#define LSR_RDR		0x01	/* Receiver Data Ready */
#define LSR_OE		0x02	/* Overrun Error */
#define LSR_PE		0x04	/* Parity Error */
#define LSR_FE		0x08	/* Framing Error */
#define LSR_BI		0x10	/* Break Interrupt */
#define LSR_THRE	0x20	/* Transmitter Holding Register Empty */
#define LSR_TEMT	0x40	/* Transmitter Empty */
#define LSR_RXFE	0x80	/* Error in RX FIFO */

/*
 * functions
 */
int uart1_init(u_long baudrate);
short peek_recv(void);
void set_to_send(u_char c);
static void uart1_handler(void) __attribute__ ((interrupt("IRQ")));

/*
 * valiables
 */
/*
 * uart1 using FIFO
 */
#define N_BUFF  16		/* must be 2^x */
#define BUFF_PTR_MASK  (N_BUFF-1)

u_char send_buff[N_BUFF];
u_long send_in_ptr, send_out_ptr;
u_long n_send;
int ti1_will_come;

u_char recv_buff[N_BUFF];
u_long recv_in_ptr, recv_out_ptr;
u_long n_recv;

/*
 *
 */
int uart1_init(u_long baudrate)
{
	u_long Fdiv;

#if USE_UART1
	PINSEL0 |= 0x00050000;	/* Enable RxD1 and TxD1 */
#else /* USE_UART1 */
	PINSEL0 |= 0x00000005;	/* Enable RxD0 and TxD0 */
#endif /* USE_UART1 */

	UxLCR = 0x83;				/* 8 bits, no Parity, 1 Stop bit */
	Fdiv = (Fpclk / 16) / baudrate;	/* calc division from baud rate */
	UxDLM = Fdiv / 256;
	UxDLL = Fdiv % 256;
	UxLCR = 0x03;               /* DLAB = 0 */

//	UxFCR = 0x07;		/* Enable and reset TX and RX FIFO. */

	/*
	 * initial both FIFO
	 */
	send_in_ptr = send_out_ptr = 0;
	n_send = 0;
	ti1_will_come = FALSE;		/* no TI1 will come */

	recv_in_ptr = recv_out_ptr = 0;
	n_recv = 0;

	/*
	 *
	 */
	if(install_irq(UARTx_INT, (void *)uart1_handler) == FALSE) {
		return(FALSE);
	}

//	UxIER = IER_RBR | IER_THRE | IER_RLS;	/* Enable UART0 interrupt */
	UxIER = IER_RBR | IER_THRE;		/* enable interrupts */

	return(TRUE);
}

/*
 *
 */
short peek_recv(void)
{
	BYTE c;
	u_long imf;

	if(n_recv == 0)
		return(-1);		/* no data has come */

//	UxIER = 0;						/* disable interrupts */
//	VICIntEnClr = 1 << UARTx_INT;	/* Disable Interrupt */
	imf = IntDisable();				/* disable interrupts */

	c = recv_buff[recv_out_ptr++];
	n_recv--;
	recv_out_ptr &= BUFF_PTR_MASK;

//	UxIER = IER_RBR | IER_THRE;		/* enable interrupts */
//	VICIntEnable = 1 << UARTx_INT;	/* Enable Interrupt */
	IntRestore(imf);				/* enable interrupts */

	return(c);
}

/*
 *
 */
void set_to_send(u_char c)
{
	u_long imf;

	/*
	 * wait if there is no space in send buffer,
	 * i.e, keep host waiting on process of SET_REPORT request
	 */
	while(n_send == N_BUFF)
		;

//	UxIER = 0;						/* disable interrupts */
//	VICIntEnClr = 1 << UARTx_INT;	/* Disable Interrupt */
	imf = IntDisable();				/* disable interrupts */

	if(ti1_will_come) {
		/*
		 * I am having transfer data, so TI1 will inform next timing.
		 */
		send_buff[send_in_ptr++] = c;
		n_send++;
		send_in_ptr &= BUFF_PTR_MASK;
if(USE_UART1)cm_putchar('s');
	} else {
		/*
		 * transmitter is idle, no TI1 will come
		 */
		UxTHR = c;			/* send directly here */
		ti1_will_come = TRUE;
if(USE_UART1)cm_putchar('d');
	}

//	UxIER = IER_RBR | IER_THRE;		/* enable interrupts */
//	VICIntEnable = 1 << UARTx_INT;	/* Enable Interrupt */
	IntRestore(imf);				/* enable interrupts */

	return;
}

/*
 *
 */
static void uart1_handler(void)
{
	BYTE iid;
	BYTE c;

	iid = UxIIR & (0x07<<1);	/* take out interrupt identification */

	if(iid == IIR_RDA) {
if(USE_UART1)cm_putchar('R');
		/*
		 * Receive Data Available
		 */
		c = UxRBR;		/* this clears interrupt request */

		if(n_recv == N_BUFF) {
			/*
			 * buffer overflow, only ignore
			 */
if(USE_UART1)cm_putchar('O');
		} else {
			recv_buff[recv_in_ptr++] = c;
			n_recv++;
			recv_in_ptr &= BUFF_PTR_MASK;
		}

		VICVectAddr = 0;		/* Acknowledge Interrupt */
		return;
	}

	if(iid == IIR_CTI) {
if(USE_UART1)cm_putchar('C');
		/*
		 * Character Time-out indicator
		 */
		/* shall never come, (do same as IIR_RDA) */

		VICVectAddr = 0;		/* Acknowledge Interrupt */
		return;
	}

	if(iid == IIR_THRE) {
if(USE_UART1)cm_putchar('T');
		/*
		 * THRE, transmit holding register empty
		 *   the request is automatically cleared when reading UxIIR
		 */
		/*
		 * Check status in the LSR to see if valid data in UxTHR or not
		 */
		if(UxLSR & LSR_THRE) {
			if(n_send == 0) {
				/*
				 * no data to send, so no TI1 will come
				 */
				ti1_will_come = FALSE;
			} else {
				UxTHR = send_buff[send_out_ptr++];
				n_send--;
				send_out_ptr &= BUFF_PTR_MASK;
			}
		}

		VICVectAddr = 0;		/* Acknowledge Interrupt */
		return;
	}

if(USE_UART1)cm_putchar('X');
	while(1)
		;

	VICVectAddr = 0;		/* Acknowledge Interrupt */
}
